# MRT Beta 1.4 - updated User Interface to support settings modifications through the settings tab
# melvin.terechenok@gmail.com

# MRT 935G BETA1.3 - Update to handle 935G
# MRT Mapped most Wouxon CPS settings in Chirp
# MRT Upload to Radio and Download from Radio are working
# MRT Settings are available under the Browser tab in Hex/Dec/Bin format
# MRT No limits or error checking is done on values entered
# MRT Incorrect/Invalid values may cause radio malfunctions



# Copyright 2019 Pavel Milanes CO7WT <pavelmc@gmail.com>
#
# Based on the work of Krystian Struzik <toner_82@tlen.pl>
# who figured out the crypt used and made possible the
# Wuoxun KG-UV8D Plus driver, in which this work is based.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Wouxun KG-935G radio management module"""

import time
import os
import logging
from chirp import util, chirp_common, bitwise, memmap, errors, directory
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    RadioSettingValueInteger, RadioSettingValueString, \
    RadioSettings


LOG = logging.getLogger(__name__)

CMD_ID = 128    # \x80
CMD_END = 129   # \x81
CMD_RD = 130    # \82
CMD_WR = 131    # \83

MEM_VALID = 158

AB_LIST = ["A", "B"]
STEPS = [2.5, 5.0, 6.25, 10.0, 12.5, 25.0, 50.0, 100.0]
STEP_LIST = [str(x) for x in STEPS]
ROGER_LIST = ["Off", "Begin", "End", "Both"]
TIMEOUT_LIST = ["Off"] + [str(x) + "s" for x in range(15, 901, 15)]
VOX_LIST = ["Off"] + ["%s" % x for x in range(1, 10)]
BANDWIDTH_LIST = ["Narrow", "Wide"]
VOICE_LIST = ["Off", "On"]
LANGUAGE_LIST = ["Chinese", "English"]
SCANMODE_LIST = ["TO", "CO", "SE"]
#MRT - EDIT FOR 935G
PFKEYLONG_LIST = ["undef", "FRQ2-PTT", "Selec Call", "Scan", "Flashlight", "Alarm", "SOS", "FM Radio", "Moni", "Strobe", "Weather", "Tlk A", "Reverse", "CTC Scan", "DCS Scan", "BRT"]
PFKEYSHORT_LIST = ["undef", "Scan", "Flashlight", "Alarm", "SOS", "FM Radio", "Moni", "Strobe", "Weather", "Tlk A", "Reverse", "CTC Scan", "DCS Scan", "BRT"]
#
WORKMODE_LIST = ["VFO", "Ch.Number.", "Ch.Freq.", "Ch.Name"]
BACKLIGHT_LIST = ["Always On"] + [str(x) + "s" for x in range(1, 21)] + \
    ["Always Off"]
OFFSET_LIST = ["+", "-"]
PONMSG_LIST = ["MSG - Bitmap", "Battery Volts"]
SPMUTE_LIST = ["QT", "QT+DTMF", "QT*DTMF"]
DTMFST_LIST = ["OFF", "DTMF", "ANI", "DTMF+ANI"]
DTMF_TIMES = [str(x) + "ms" for x in range(0, 501, 10)]
RPTSET_LIST = ["", "X-DIRRPT", "X-TWRPT"] # TODO < what is index 0?
ALERTS = [1750, 2100, 1000, 1450]
ALERTS_LIST = [str(x) for x in ALERTS]
PTTID_LIST = ["BOT", "EOT", "Both"]
LIST_10 = ["Off"] + ["%s" % x for x in range(1, 11)]
SCANGRP_LIST = ["All"] + ["%s" % x for x in range(1, 11)]
SCQT_LIST = ["Decoder", "Encoder", "All"]
SMUTESET_LIST = ["Off", "Tx", "Rx", "Tx+Rx"]
POWER_LIST = ["Lo", "Mid", "Hi"]
HOLD_TIMES = ["Off"] + ["%s" % x for x in range(100, 5001, 100)]
RPTMODE_LIST = ["Radio", "Repeater"]
#MRT ADDED NEW LISTS
THEME_LIST = ["White-1", "White-2", "Black-1", "Black-2"]
DSPBRTSBY_LIST = ["OFF", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"]
DSPBRTACT_LIST = ["OFF-DO NOT USE", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"]
TONESCANSAVELIST= ["Rx", "Tx", "Tx/Rx"]
PTTDELAY_LIST = [str(x) + "ms" for x in range(0, 3001, 100)]
SCRAMBLE_LIST = ["OFF"] + [str(x) for x in range(1,9)]

# memory slot 0 is not used, start at 1 (so need 1000 slots, not 999)
# structure elements whose name starts with x are currently unidentified

# MRT made Power = 2 bits to handle 935G's 3 power levels
# MRT beta 1.3 - updates to structure to match KG935G Custom programming SW configuration settings, FM Radio presets, Key Settings
# MRT beta 1.4 - modified some value names and merged FM Radio memories with other settings to resolve errors in UI implementation
_MEM_FORMAT = """
    #seekto 0x0044;
    struct {
        u32    rx_start;
        u32    rx_stop;
        u32    tx_start;
        u32    tx_stop;
    } uhf_limits;

    #seekto 0x0054;
    struct {
        u32    rx_start;
        u32    rx_stop;
        u32    tx_start;
        u32    tx_stop;
    } vhf_limits;

    #seekto 0x0400;
    struct {
        char     oem1[8];
        char     unknown[2];
        u8     unknown2[10];
        u8     unknown3[10];
        u8     unknown4[8];
        char     oem2[10];
        u8     version[6];
        char     date[8];
        u8     unknown5[2];
        char     model[8];     
    } oem_info;

    #seekto 0x0480;
    struct {
        u16    Group_1_lower;
        u16    Group_1_upper;
        u16    Group_2_lower;
        u16    Group_2_upper;
        u16    Group_3_lower;
        u16    Group_3_upper;
        u16    Group_4_lower;
        u16    Group_4_upper;
        u16    Group_5_lower;
        u16    Group_5_upper;
        u16    Group_6_lower;
        u16    Group_6_upper;
        u16    Group_7_lower;
        u16    Group_7_upper;
        u16    Group_8_lower;
        u16    Group_8_upper;
        u16    Group_9_lower;
        u16    Group_9_upper;
        u16    Group_10_lower;
        u16    Group_10_upper;
    } scan_groups;

    #seekto 0x0500;
    struct {
        u8    call_code_1[6];
        u8    call_code_2[6];
        u8    call_code_3[6];
        u8    call_code_4[6];
        u8    call_code_5[6];
        u8    call_code_6[6];
        u8    call_code_7[6];
        u8    call_code_8[6];
        u8    call_code_9[6];
        u8    call_code_10[6];
        u8    call_code_11[6];
        u8    call_code_12[6];
        u8    call_code_13[6];
        u8    call_code_14[6];
        u8    call_code_15[6];
        u8    call_code_16[6];
        u8    call_code_17[6];
        u8    call_code_18[6];
        u8    call_code_19[6];
        u8    call_code_20[6];
    } call_groups;

    #seekto 0x0580;
    struct {
        char    call_name1[6];
        char    call_name2[6];
        char    call_name3[6];
        char    call_name4[6];
        char    call_name5[6];
        char    call_name6[6];
        char    call_name7[6];
        char    call_name8[6];
        char    call_name9[6];
        char    call_name10[6];
        char    call_name11[6];
        char    call_name12[6];
        char    call_name13[6];
        char    call_name14[6];
        char    call_name15[6];
        char    call_name16[6];
        char    call_name17[6];
        char    call_name18[6];
        char    call_name19[6];
        char    call_name20[6];
    } call_names;


    #seekto 0x0600;
    struct {
        u16    FM_radio1;
        u16    FM_radio2;
        u16    FM_radio3;
        u16    FM_radio4;
        u16    FM_radio5;
        u16    FM_radio6;
        u16    FM_radio7;
        u16    FM_radio8;
        u16    FM_radio9;
        u16    FM_radio10;
        u16    FM_radio11;
        u16    FM_radio12;
        u16    FM_radio13;
        u16    FM_radio14;
        u16    FM_radio15;
        u16    FM_radio16;
        u16    FM_radio17;
        u16    FM_radio18;
        u16    FM_radio19;
        u16    FM_radio20;
        u16 unknown_pad_x0640[235];
        u8 unknown07fe;
        u8 unknown07ff;
        u8      ponmsg;
        char    dispstr[15];
        u8 unknown0810;
        u8 unknown0811;
        u8 unknown0812;
        u8 unknown0813;
        u8 unknown0814;
        u8      voice;
        u8      timeout;
        u8      toalarm;
        u8      channel_menu;
        u8      power_save;
        u8      autolock;
        u8      keylock;
        u8      beep;
        u8      stopwatch;
        u8      vox;
        u8      scan_rev;
        u8      backlight;
        u8      roger_beep;
        char      node_sw_pwd[6];
        char      reset_pwd[6];
        u8 unknown082e;
        u8      pri_ch;
        u8      ani_sw;
        u8      ptt_delay;
        u8      ani_code[6];
        u8      dtmf_st;
        u8      BCL_A;
        u8      BCL_B;
        u8      ptt_id;
        u8      prich_sw;
        u8 unknown083d;
        u8 unknown083e;
        u8 unknown083f;
        u8      alert;
        u8      pf1_shrt;
        u8      pf1_long;
        u8      pf2_shrt;
        u8      pf2_long;
        u8 unknown0845;
        u8      work_mode_a;
        u8      work_mode_b;
        u8      dtmf_tx_time;
        u8      dtmf_interval;
        u8      main_band;
        u8 unknown084b;
        u8      work_ch_a;
        u8 unknown084d;
        u8      work_ch_b;
        u8 unknown084f;
        u8 unknown0850;
        u8 unknown0851;
        u8 unknown0852;
        u8 unknown0853;
        u8 unknown0854;
        u8 unknown0855;
        u8 unknown0856;
        u8 unknown0857;
        u8 unknown0858;
        u8 unknown0859;
        u8 unknown085a;
        u8 unknown085b;
        u8 unknown085c;
        u8 unknown085d;
        u8 unknown085e;
        u8 unknown085f;
        u8 unknown0860;
        u8      TDR_single_mode;
        u8      ring_time;
        u8      ScnGrpA_Act;
        u8      ScnGrpB_Act;
        u8 unknown0865;
        u8      rpt_tone;
        u8 unknown0867;
        u8      scan_det;
        u8      ToneScnSave;
        u8 unknown086a;
        u8      smuteset;
        u8 unknown086c;
        u8      DspBrtAct;
        u8      DspBrtSby;
        u8 unknown086f;
        u8      theme;
        u8      wxalert;
        u8      VFO_repeater_a;
        u8      VFO_repeater_b;
        u8 unknown0874;
        u8 unknown0875;
        u8 unknown0876;
        u8 unknown0877;
        u8 unknown0878;
        u8 unknown0879;
        u8 unknown087a;
        u8 unknown087b;
        u8 unknown087c;
        u8 unknown087d;
        u8 unknown087e;
        u8 unknown087f;
    } settings;

    #seekto 0x0880;
    struct {
        u32     rxfreq;
        u32     unknown0;
        u16     rxtone;
        u16     txtone;
        u8      scrambler:4,
                unknown1:2,
                power:2; 
        u8      unknown3:1,
                unknown5:2
                unknown4:1,
                compander:1,
                mute_mode:2,
                iswide:1;
        u8      step;
        u8      squelch;
      } vfoa;

    #seekto 0x08c0;
    struct {
        u32     rxfreq;
        u32     unknown0;
        u16     rxtone;
        u16     txtone;
        u8      scrambler:4,
                unknown1:2,
                power:2; 
        u8      unknown3:1,
                unknown5:2
                unknown4:1,
                compander:1,
                mute_mode:2,
                iswide:1;
        u8      step;
        u8      squelch;
    } vfob;

    #seekto 0x0900;
    struct {
        u32     rxfreq;
        u32     txfreq;
        u16     rxtone;
        u16     txtone;
        u8      scrambler:4,
                unknown1:2,
                power:2; 
        u8      unknown3:2,
                scan_add:1,
                unknown4:1,
                compander:1,
                mute_mode:2,
                iswide:1;
        u16     padding;
    } memory[1000];

    #seekto 0x4780;
    struct {
        u8    name[8];
                u8    unknown[4];
    } names[1000];

    #seekto 0x7670;
    u8          valid[1000];
    """

    # Support for the Wouxun KG-935G radio
    # Serial coms are at 19200 baud
    # The data is passed in variable length records
    # Record structure:
    #  Offset   Usage
    #    0      start of record (\x7c)
    #    1      Command (\x80 Identify \x81 End/Reboot \x82 Read \x83 Write)
    #    2      direction (\xff PC-> Radio, \x00 Radio -> PC)
    #    3      length of payload (excluding header/checksum) (n)
    #    4      payload (n bytes)
    #    4+n+1  checksum - byte sum (% 256) of bytes 1 -> 4+n
    #
    # Memory Read Records:
    # the payload is 3 bytes, first 2 are offset (big endian),
    # 3rd is number of bytes to read
    # Memory Write Records:
    # the maximum payload size (from the Wouxun software) seems to be 66 bytes
    #  (2 bytes location + 64 bytes data).

#MRT 1.2 correct spelling of Wouxon
class KGUV8TRadio(chirp_common.Alias):
    VENDOR = "Wouxun"
    MODEL = "KG-935G"

@directory.register
class KG935GRadio(chirp_common.CloneModeRadio,
                  chirp_common.ExperimentalRadio):

    """Wouxun KG-935G"""
    VENDOR = "Wouxun"
    MODEL = "KG-935G"
    _model = "KG-UV8D-B"
    _file_ident = "935G"
    BAUD_RATE = 19200
# MRT - Added Medium Power level for 935G support
    POWER_LEVELS = [chirp_common.PowerLevel("L", watts=0.5),
                    chirp_common.PowerLevel("M", watts=4.5),
                    chirp_common.PowerLevel("H", watts=5.5)]
    _mmap = ""
    ALIASES = [KGUV8TRadio,]

    def _checksum(self, data):
        cs = 0
        for byte in data:
            cs += ord(byte)
        return chr(cs % 256)

    def _write_record(self, cmd, payload = None):
        # build the packet
        _header = '\x7c' + chr(cmd) + '\xff'

        _length = 0
        if payload:
            _length = len(payload)

        # update the length field
        _header += chr(_length)

        if payload:
            # calculate checksum then add it with the payload to the packet and encrypt
            crc = self._checksum(_header[1:] + payload)
            payload += crc
            _header += self.encrypt(payload)
        else:
            # calculate and add encrypted checksum to the packet
            crc = self._checksum(_header[1:])
            _header += self.strxor(crc, '\x57')

        try:
            self.pipe.write(_header)
        except Exception, e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)

    def _read_record(self):
        # read 4 chars for the header
        _header = self.pipe.read(4)
        if len(_header) != 4:
            raise errors.RadioError('Radio did not respond')
        _length = ord(_header[3])
        _packet = self.pipe.read(_length)
        _rcs_xor = _packet[-1]
        _packet = self.decrypt(_packet)
        _cs = ord(self._checksum(_header[1:] + _packet))
        # read the checksum and decrypt it
        _rcs = ord(self.strxor(self.pipe.read(1), _rcs_xor))
        return (_rcs != _cs, _packet)

    def decrypt(self, data):
        result = ''
        for i in range(len(data)-1, 0, -1):
            result += self.strxor(data[i], data[i - 1])
        result += self.strxor(data[0], '\x57')
        return result[::-1]

    def encrypt(self, data):
        result = self.strxor('\x57', data[0])
        for i in range(1, len(data), 1):
            result += self.strxor(result[i - 1], data[i])
        return result

    def strxor (self, xora, xorb):
        return chr(ord(xora) ^ ord(xorb))

    # Identify the radio
    #
    # A Gotcha: the first identify packet returns a bad checksum, subsequent
    # attempts return the correct checksum... (well it does on my radio!)
    #
    # The ID record returned by the radio also includes the current frequency range
    # as 4 bytes big-endian in 10Hz increments
    #
    # Offset
    #  0:10     Model, zero padded (Looks for 'KG-UV8D-B')

    @classmethod
    def match_model(cls, filedata, filename):
        id = cls._file_ident 
        return cls._file_ident in filedata[0x426:0x430]

    def _identify(self):
        """Do the identification dance"""
        for _i in range(0, 10):
            self._write_record(CMD_ID)
            _chksum_err, _resp = self._read_record()
            LOG.debug("Got:\n%s" % util.hexprint(_resp))
            if _chksum_err:
                LOG.error("Checksum error: retrying ident...")
                time.sleep(0.100)
                continue
            LOG.debug("Model %s" % util.hexprint(_resp[0:9]))
            if _resp[0:9] == self._model:
                return
            if len(_resp) == 0:
                raise Exception("Radio not responding")
            else:
                raise Exception("Unable to identify radio")

    def _finish(self):
        self._write_record(CMD_END)

    def process_mmap(self):
        self._memobj = bitwise.parse(_MEM_FORMAT, self._mmap)

    def sync_in(self):
        try:
            self._mmap = self._download()
        except errors.RadioError:
            raise
        except Exception, e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        self.process_mmap()

    def sync_out(self):
        self._upload()

    # TODO: Load all memory.
    # It would be smarter to only load the active areas and none of
    # the padding/unused areas. Padding still need to be investigated.
    def _download(self):
        """Talk to a wouxun KG-935G and do a download"""
        try:
            self._identify()
            return self._do_download(0, 32768, 64)
        except errors.RadioError:
            raise
        except Exception, e:
            LOG.exception('Unknown error during download process')
            raise errors.RadioError("Failed to communicate with radio: %s" % e)

    def _do_download(self, start, end, blocksize):
        # allocate & fill memory
        image = ""
        for i in range(start, end, blocksize):
            req = chr(i / 256) + chr(i % 256) + chr(blocksize)
            self._write_record(CMD_RD, req)
            cs_error, resp = self._read_record()
            if cs_error:
                LOG.debug(util.hexprint(resp))
                raise Exception("Checksum error on read")
            # LOG.debug("Got:\n%s" % util.hexprint(resp))
            image += resp[2:]
            if self.status_fn:
                status = chirp_common.Status()
                status.cur = i
                status.max = end
                status.msg = "Cloning from radio"
                self.status_fn(status)
        self._finish()
        return memmap.MemoryMap(''.join(image))

    def _upload(self):
        """Talk to a wouxun KG-935G and do a upload"""
        try:
            self._identify()
            self._do_upload(0, 32768, 64)
        except errors.RadioError:
            raise
        except Exception, e:
            raise errors.RadioError("Failed to communicate with radio: %s" % e)
        return

    def _do_upload(self, start, end, blocksize):
        ptr = start
        for i in range(start, end, blocksize):
            req = chr(i / 256) + chr(i % 256)
            chunk = self.get_mmap()[ptr:ptr + blocksize]
            self._write_record(CMD_WR, req + chunk)
            LOG.debug(util.hexprint(req + chunk))
            cserr, ack = self._read_record()
            LOG.debug(util.hexprint(ack))
            j = ord(ack[0]) * 256 + ord(ack[1])
            if cserr or j != ptr:
                raise Exception("Radio did not ack block %i" % ptr)
            ptr += blocksize
            if self.status_fn:
                status = chirp_common.Status()
                status.cur = i
                status.max = end
                status.msg = "Cloning to radio"
                self.status_fn(status)
        self._finish()

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_ctone = True
        rf.has_rx_dtcs = True
        rf.has_cross = True
        rf.has_tuning_step = False
        rf.has_bank = False
        rf.can_odd_split = True
        rf.valid_skips = ["", "S"]
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_cross_modes = [
            "Tone->Tone",
            "Tone->DTCS",
            "DTCS->Tone",
            "DTCS->",
            "->Tone",
            "->DTCS",
            "DTCS->DTCS",
        ]
        rf.valid_modes = ["FM", "NFM"]
        rf.valid_power_levels = self.POWER_LEVELS
        rf.valid_name_length = 8
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_bands = [(137000000, 175000000),  # supports 2m
                          (400000000, 480000000)]  # supports 70cm
        rf.valid_characters = chirp_common.CHARSET_ASCII
        rf.memory_bounds = (1, 999)  # 999 memories
        rf.valid_tuning_steps = STEPS
        return rf

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = \
            ('This driver is experimental.\n'
             '\n'
             'Please keep a copy of your memories with the original Wouxon CPS software '
             'if you treasure them, this driver is new and may contain bugs.\n'
             '\n'
             ' Upload and Download from/to the Radio are working\n'
             ' '
             ' User interface has been added. \n Most of the settings from the Wouxon Custom Programming'
             ' software have been mapped\n in addition to some bonus settings that were'
             ' found.\n'
             ' Changing the VHF/UHF Rx limits does appear to work - but radio performance\n'
             ' is not guaranteed-  and may void warranty or cause radio to malfunction.\n'
             ' You can also customize the bottom banner from the OEMINFO Model setting\n'
                          ' \n'
             ' There are no limits/error checking done on the settings when using the Browser tab.\n'
             ' \n'
             ' Incorrect/Illegal values for a setting may cause radio malfunctions\n'
             ' USE AT YOUR OWN RISK\n'
             '\n'
             )
        return rp

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])
# MRT - corrected the Polarity decoding to match 935G implementation use 0x2000 bit mask for R
# MRT - 0x2000 appears to be the bit mask for Inverted DCS tones
# MRT - n DCS Tone will be 0x4xxx values - i DCS Tones will be 0x6xxx values.
# MRT - Chirp Uses N for n DCS Tones and R for i DCS Tones
    def _get_tone(self, _mem, mem):
        def _get_dcs(val):
            code = int("%03o" % (val & 0x07FF))
            pol = (val & 0x2000) and "R" or "N"
            return code, pol
# MRT - Modified the function below to bitwise AND with 0x4000 to check for 935G DCS Tone decoding
# MRT 0x4000 appears to be the bit mask for DCS tones
        tpol = False
# MRT Beta 1.1 - Fix the txtone compare to 0x4000 - was rxtone.
        if _mem.txtone != 0xFFFF and (_mem.txtone & 0x4000) == 0x4000:
            tcode, tpol = _get_dcs(_mem.txtone)
            mem.dtcs = tcode
            txmode = "DTCS"
        elif _mem.txtone != 0xFFFF and _mem.txtone != 0x0:
            mem.rtone = (_mem.txtone & 0x7fff) / 10.0
            txmode = "Tone"
        else:
            txmode = ""
# MRT - Modified the function below to bitwise AND with 0x4000 to check for 935G DCS Tone decoding
        rpol = False
        if _mem.rxtone != 0xFFFF and (_mem.rxtone & 0x4000) == 0x4000:
            rcode, rpol = _get_dcs(_mem.rxtone)
            mem.rx_dtcs = rcode
            rxmode = "DTCS"
        elif _mem.rxtone != 0xFFFF and _mem.rxtone != 0x0:
            mem.ctone = (_mem.rxtone & 0x7fff) / 10.0
            rxmode = "Tone"
        else:
            rxmode = ""

        if txmode == "Tone" and not rxmode:
            mem.tmode = "Tone"
        elif txmode == rxmode and txmode == "Tone" and mem.rtone == mem.ctone:
            mem.tmode = "TSQL"
        elif txmode == rxmode and txmode == "DTCS" and mem.dtcs == mem.rx_dtcs:
            mem.tmode = "DTCS"
        elif rxmode or txmode:
            mem.tmode = "Cross"
            mem.cross_mode = "%s->%s" % (txmode, rxmode)

        # always set it even if no dtcs is used
        mem.dtcs_polarity = "%s%s" % (tpol or "N", rpol or "N")

        LOG.debug("Got TX %s (%i) RX %s (%i)" %
                  (txmode, _mem.txtone, rxmode, _mem.rxtone))

    def get_memory(self, number):
        _mem = self._memobj.memory[number]
        _nam = self._memobj.names[number]

        mem = chirp_common.Memory()
        mem.number = number
        _valid = self._memobj.valid[mem.number]
        LOG.debug("%d %s", number, _valid == MEM_VALID)
        if _valid != MEM_VALID:
            mem.empty = True
            return mem
        else:
            mem.empty = False

        mem.freq = int(_mem.rxfreq) * 10

        if _mem.txfreq == 0xFFFFFFFF:
            # TX freq not set
            mem.duplex = "off"
            mem.offset = 0
        elif int(_mem.rxfreq) == int(_mem.txfreq):
            mem.duplex = ""
            mem.offset = 0
        elif abs(int(_mem.rxfreq) * 10 - int(_mem.txfreq) * 10) > 70000000:
            mem.duplex = "split"
            mem.offset = int(_mem.txfreq) * 10
        else:
            mem.duplex = int(_mem.rxfreq) > int(_mem.txfreq) and "-" or "+"
            mem.offset = abs(int(_mem.rxfreq) - int(_mem.txfreq)) * 10

        for char in _nam.name:
            if char != 0:
                mem.name += chr(char)
        mem.name = mem.name.rstrip()

        self._get_tone(_mem, mem)

        mem.skip = "" if bool(_mem.scan_add) else "S"

        mem.power = self.POWER_LEVELS[_mem.power]
        mem.mode = _mem.iswide and "FM" or "NFM"
        return mem

    def _set_tone(self, mem, _mem):
        def _set_dcs(code, pol):
#MRT Change from + 0x2800 to bitwise OR with 0x4000 to set the bit for DCS
            val = int("%i" % code, 8) | 0x4000
            if pol == "R":
#MRT Change to 0x2000 from 0x8000 to set the bit for i/R polarity
               val += 0x2000
            return val

        rx_mode = tx_mode = None
        rxtone = txtone = 0x0000

        if mem.tmode == "Tone":
            tx_mode = "Tone"
            rx_mode = None
            txtone = int(mem.rtone * 10) + 0x8000
        elif mem.tmode == "TSQL":
            rx_mode = tx_mode = "Tone"
            rxtone = txtone = int(mem.ctone * 10) + 0x8000
        elif mem.tmode == "DTCS":
            tx_mode = rx_mode = "DTCS"
            txtone = _set_dcs(mem.dtcs, mem.dtcs_polarity[0])
            rxtone = _set_dcs(mem.dtcs, mem.dtcs_polarity[1])
        elif mem.tmode == "Cross":
            tx_mode, rx_mode = mem.cross_mode.split("->")
            if tx_mode == "DTCS":
                txtone = _set_dcs(mem.dtcs, mem.dtcs_polarity[0])
            elif tx_mode == "Tone":
                txtone = int(mem.rtone * 10) + 0x8000
            if rx_mode == "DTCS":
                rxtone = _set_dcs(mem.rx_dtcs, mem.dtcs_polarity[1])
            elif rx_mode == "Tone":
                rxtone = int(mem.ctone * 10) + 0x8000

        _mem.rxtone = rxtone
        _mem.txtone = txtone

        LOG.debug("Set TX %s (%i) RX %s (%i)" %
                  (tx_mode, _mem.txtone, rx_mode, _mem.rxtone))

    def set_memory(self, mem):
        number = mem.number

        _mem = self._memobj.memory[number]
        _nam = self._memobj.names[number]

        if mem.empty:
            _mem.set_raw("\x00" * (_mem.size() / 8))
            self._memobj.valid[number] = 0
            self._memobj.names[number].set_raw("\x00" * (_nam.size() / 8))
            return

        _mem.rxfreq = int(mem.freq / 10)
        if mem.duplex == "off":
            _mem.txfreq = 0xFFFFFFFF
        elif mem.duplex == "split":
            _mem.txfreq = int(mem.offset / 10)
        elif mem.duplex == "off":
            for i in range(0, 4):
                _mem.txfreq[i].set_raw("\xFF")
        elif mem.duplex == "+":
            _mem.txfreq = int(mem.freq / 10) + int(mem.offset / 10)
        elif mem.duplex == "-":
            _mem.txfreq = int(mem.freq / 10) - int(mem.offset / 10)
        else:
            _mem.txfreq = int(mem.freq / 10)
        _mem.scan_add = int(mem.skip != "S")
        _mem.iswide = int(mem.mode == "FM")
        # set the tone
        self._set_tone(mem, _mem)
        # set the scrambler and compander to off by default
        _mem.scrambler = 0
        _mem.compander = 0
        # set the power
        if mem.power:
            _mem.power = self.POWER_LEVELS.index(mem.power)
        else:
            _mem.power = True
        # set to mute mode to QT (not QT+DTMF or QT*DTMF) by default
        _mem.mute_mode = 0

        for i in range(0, len(_nam.name)):
            if i < len(mem.name) and mem.name[i]:
                _nam.name[i] = ord(mem.name[i])
            else:
                _nam.name[i] = 0x0
        self._memobj.valid[mem.number] = MEM_VALID

    def _get_settings(self):
        _settings = self._memobj.settings
        _vfoa = self._memobj.vfoa
        _vfob = self._memobj.vfob
        _scan = self._memobj.scan_groups
        _call = self._memobj.call_groups
        _callname = self._memobj.call_names

        cfg_grp = RadioSettingGroup("cfg_grp", "Config Settings")
        vfoa_grp = RadioSettingGroup("vfoa_grp", "VFO A Settings")
        vfob_grp = RadioSettingGroup("vfob_grp", "VFO B Settings")
        key_grp = RadioSettingGroup("key_grp", "Key Settings")
        fmradio_grp = RadioSettingGroup("fmradio_grp", "FM Broadcast Memory")
        lmt_grp = RadioSettingGroup("lmt_grp", "Rx Frequency Limits")
        uhf_lmt_grp = RadioSettingGroup("uhf_lmt_grp", "UHF")
        vhf_lmt_grp = RadioSettingGroup("vhf_lmt_grp", "VHF")
        oem_grp = RadioSettingGroup("oem_grp", "OEM Info")
        scan_grp = RadioSettingGroup("scan_grp", "Scan Group")
        call_grp = RadioSettingGroup("call_grp", "Call Settings")
        extra_grp = RadioSettingGroup("extra_grp", "NOT CHANGED BY RESET OR CPS")
        # lmt_grp.append(vhf_lmt_grp);
        # lmt_grp.append(uhf_lmt_grp);
        extra_grp.append(lmt_grp);
        extra_grp.append(oem_grp);
        group = RadioSettings(cfg_grp, vfoa_grp, vfob_grp,
                             fmradio_grp, key_grp, scan_grp, call_grp, extra_grp)


    # Call Settings

        callchars = "0123456789"
 
        # # callnamechars = chirp_common.CHARSET_ASCII
        # _name = str(_callname.call_name1).split("\0")[0]
        # # if any((c in callnamechars) for c in _msg):
        # val = RadioSettingValueString(0, 6, _name)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name1", "Call Name 1", val)
        # call_grp.append(rs)
        # # else:

        _codeobj = _call.call_code_1
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_1", "Call Code 1", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_1)
        call_grp.append(rs)

        # _msg = str(_callname.call_name2).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name2", "Call Name 2", val)
        # call_grp.append(rs)

        _codeobj = _call.call_code_2
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_2", "Call Code 2", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_2)
        call_grp.append(rs)

        # _msg = str(_callname.call_name3).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name3", "Call Name 3", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_3
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_3", "Call Code 3", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_3)
        call_grp.append(rs)

        # _msg = str(_callname.call_name4).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name4", "Call Name 4", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_4
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_4", "Call Code 4", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_4)
        call_grp.append(rs)

        # _msg = str(_callname.call_name5).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name5", "Call Name 5", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_5
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_5", "Call Code 5", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_5)
        call_grp.append(rs)

        # _msg = str(_callname.call_name6).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name6", "Call Name 6", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_6
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_6", "Call Code 6", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_6)
        call_grp.append(rs)

        # _msg = str(_callname.call_name7).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name7", "Call Name 7", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_7
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_7", "Call Code 7", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_7)
        call_grp.append(rs)

        # _msg = str(_callname.call_name8).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name8", "Call Name 8", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_8
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_8", "Call Code 8", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_8)
        call_grp.append(rs)

        # _msg = str(_callname.call_name9).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name9", "Call Name 9", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_9
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_9", "Call Code 9", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_9)
        call_grp.append(rs)

        # _msg = str(_callname.call_name10).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name10", "Call Name 10", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_10
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_10", "Call Code 10", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_10)
        call_grp.append(rs)

        # _msg = str(_callname.call_name11).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name11", "Call Name 11", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_11
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_11", "Call Code 11", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_11)
        call_grp.append(rs)

        # _msg = str(_callname.call_name12).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name12", "Call Name 12", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_12
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_12", "Call Code 12", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_12)
        call_grp.append(rs)

        # _msg = str(_callname.call_name13).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name13", "Call Name 13", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_13
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_13", "Call Code 13", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_13)
        call_grp.append(rs)

        # _msg = str(_callname.call_name14).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name14", "Call Name 14", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_14
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_14", "Call Code 14", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_14)
        call_grp.append(rs)

        # _msg = str(_callname.call_name15).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name15", "Call Name 15", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_15
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_15", "Call Code 15", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_15)
        call_grp.append(rs)

        # _msg = str(_callname.call_name16).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name16", "Call Name 16", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_16
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_16", "Call Code 16", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_16)
        call_grp.append(rs)

        # _msg = str(_callname.call_name17).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name17", "Call Name 17", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_17
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_17", "Call Code 17", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_17)
        call_grp.append(rs)

        # _msg = str(_callname.call_name18).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name18", "Call Name 18", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_18
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_18", "Call Code 18", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_18)
        call_grp.append(rs)

        # _msg = str(_callname.call_name19).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name19", "Call Name 19", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_19
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_19", "Call Code 19", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_19)
        call_grp.append(rs)

        # _msg = str(_callname.call_name20).split("\0")[0]
        # val = RadioSettingValueString(0, 6, _msg)
        # val.set_mutable(True)
        # rs = RadioSetting("call_name20", "Call Name 20", val)
        # call_grp.append(rs)


        _codeobj = _call.call_code_20
        _code = "".join([callchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(callchars)
        rs = RadioSetting("call_code_20", "Call Code 20", val)
        def apply_call_code(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(callchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.call_code = value
        rs.set_apply_callback(apply_call_code, _call.call_code_20)
        call_grp.append(rs)

        # Configuration Settings
        #
        rs = RadioSetting("DspBrtAct", "Display Brightnes ACTIVE",
                          RadioSettingValueList(DSPBRTACT_LIST,
                                                 DSPBRTACT_LIST[_settings.DspBrtAct]))
        cfg_grp.append(rs)
        rs = RadioSetting("DspBrtSby", "Display Brightnes STANDBY",
                          RadioSettingValueList(DSPBRTSBY_LIST,
                                                 DSPBRTSBY_LIST[_settings.DspBrtSby]))
        cfg_grp.append(rs)
        rs = RadioSetting("wxalert", "Weather Alert",
                          RadioSettingValueBoolean(_settings.wxalert))
        cfg_grp.append(rs)
        rs = RadioSetting("power_save", "Battery Saver",
                          RadioSettingValueBoolean(_settings.power_save))
        cfg_grp.append(rs)
        rs = RadioSetting("theme", "Theme",
                          RadioSettingValueList(
                              THEME_LIST, THEME_LIST[_settings.theme]))
        cfg_grp.append(rs)
        rs = RadioSetting("backlight", "Backlight Active Time",
                          RadioSettingValueList(BACKLIGHT_LIST,
                                                BACKLIGHT_LIST[_settings.
                                                               backlight]))
        cfg_grp.append(rs)
        rs = RadioSetting("scan_rev", "Scan Mode",
                          RadioSettingValueList(SCANMODE_LIST,
                                                SCANMODE_LIST[_settings.
                                                              scan_rev]))
        cfg_grp.append(rs)
        rs = RadioSetting("prich_sw", "Priority Channel Scan",
                          RadioSettingValueBoolean(_settings.prich_sw))
        cfg_grp.append(rs)
        rs = RadioSetting("pri_ch", "Priority Channel - Can not be empty Channel",
                          RadioSettingValueInteger(1, 999, _settings.pri_ch))
        cfg_grp.append(rs)
        rs = RadioSetting("scan_det", "Scan Mode Tone Detect",
                          RadioSettingValueBoolean(_settings.scan_det))
        cfg_grp.append(rs)
        rs = RadioSetting("ToneScnSave", "Tone Scan Save",
                          RadioSettingValueList(TONESCANSAVELIST,
                                                TONESCANSAVELIST[_settings.ToneScnSave]))
        cfg_grp.append(rs)
        rs = RadioSetting("roger_beep", "Roger Beep",
                          RadioSettingValueList(ROGER_LIST,
                                                ROGER_LIST[_settings.roger_beep]))
        cfg_grp.append(rs)
        rs = RadioSetting("timeout", "Timeout Timer (TOT)",
                          RadioSettingValueList(
                              TIMEOUT_LIST, TIMEOUT_LIST[_settings.timeout]))
        cfg_grp.append(rs)
        rs = RadioSetting("toalarm", "Timeout Alarm (TOA)",
                          RadioSettingValueInteger(0, 10, _settings.toalarm))
        cfg_grp.append(rs)
        rs = RadioSetting("vox", "VOX",
                          RadioSettingValueList(LIST_10,
                                                LIST_10[_settings.vox]))
        cfg_grp.append(rs)
        rs = RadioSetting("voice", "Voice Guide",
                          RadioSettingValueBoolean(_settings.voice))
        cfg_grp.append(rs)
        rs = RadioSetting("beep", "Keypad Beep",
                          RadioSettingValueBoolean(_settings.beep))
        cfg_grp.append(rs)
        rs = RadioSetting("BCL_B", "Busy Channel Lock-out A",
                          RadioSettingValueBoolean(_settings.BCL_A))
        cfg_grp.append(rs)
        rs = RadioSetting("BCL_A", "Busy Channel Lock-out B",
                          RadioSettingValueBoolean(_settings.BCL_B))
        cfg_grp.append(rs)
        rs = RadioSetting("smuteset", "Secondary Area Mute (SMUTESET)",
                          RadioSettingValueList(SMUTESET_LIST,
                                                SMUTESET_LIST[_settings.
                                                              smuteset]))
        cfg_grp.append(rs)
        rs = RadioSetting("ani_sw", "ANI-ID Switch (ANI-SW)",
                          RadioSettingValueBoolean(_settings.ani_sw))
        cfg_grp.append(rs)
        rs = RadioSetting("dtmf_st", "DTMF Sidetone (SIDETONE)",
                          RadioSettingValueList(DTMFST_LIST,
                                                DTMFST_LIST[_settings.
                                                            dtmf_st]))
        cfg_grp.append(rs)
        rs = RadioSetting("alert", "Alert Tone",
                          RadioSettingValueList(ALERTS_LIST,
                                                ALERTS_LIST[_settings.alert]))
        cfg_grp.append(rs)
        rs = RadioSetting("ptt_delay", "PTT-DLY - MIN 100ms",
                          RadioSettingValueList(PTTDELAY_LIST,
                                                PTTDELAY_LIST[_settings.ptt_delay]))
        cfg_grp.append(rs)
        rs = RadioSetting("ptt_id", "PTT-ID",
                          RadioSettingValueList(PTTID_LIST,
                                                PTTID_LIST[_settings.ptt_id]))
        cfg_grp.append(rs)
        rs = RadioSetting("ring_time", "Ring Time",
                          RadioSettingValueList(LIST_10,
                                                LIST_10[_settings.ring_time]))
        cfg_grp.append(rs)
        rs = RadioSetting("rpt_tone", "Repeater Tone",
                          RadioSettingValueBoolean(_settings.rpt_tone))
        cfg_grp.append(rs)
        rs = RadioSetting("stopwatch", "Timer / Stopwatch",
                          RadioSettingValueBoolean(_settings.stopwatch))
        cfg_grp.append(rs)
        rs = RadioSetting("autolock", "Autolock",
                          RadioSettingValueBoolean(_settings.autolock))
        cfg_grp.append(rs)
        rs = RadioSetting("keylock", "Keypad Lock",
                          RadioSettingValueBoolean(_settings.keylock))
        cfg_grp.append(rs)
        rs = RadioSetting("ponmsg", "Poweron message",
                          RadioSettingValueList(
                              PONMSG_LIST, PONMSG_LIST[_settings.ponmsg]))
        cfg_grp.append(rs)
        rs = RadioSetting("dtmf_tx_time", "DTMF Transmit Time - MIN 50ms",
                          RadioSettingValueList(DTMF_TIMES,
                                                DTMF_TIMES[_settings.
                                                           dtmf_tx_time]))
        cfg_grp.append(rs)
        rs = RadioSetting("dtmf_interval", "DTMF Interval Time - MIN 50ms",
                          RadioSettingValueList(DTMF_TIMES,
                                                DTMF_TIMES[_settings.
                                                           dtmf_interval]))
        cfg_grp.append(rs)
        rs = RadioSetting("channel_menu", "Menu available in channel mode",
                          RadioSettingValueBoolean(_settings.channel_menu))
        cfg_grp.append(rs)

        # pswdchars = "0123456789"
        # _codeobj = _settings.node_sw_pwd
        # _code = "".join([pswdchars[x] for x in _codeobj if int(x) < 0x0A])
        # val = RadioSettingValueString(3, 6, _code, False)
        # val.set_charset(pswdchars)
        # rs = RadioSetting("node_sw_pwd", "Node SW Password", val)
        # def apply_node_pwd(setting, obj):
        #     value = []
        #     for j in range(0, 6):
        #         try:
        #             value.append(pswdchars.index(str(setting.value)[j]))
        #         except IndexError:
        #             value.append(0xFF)
        #     obj.node_sw_pwd = value
        # rs.set_apply_callback(apply_node_pwd, _settings)
        # key_grp.append(rs)

        # pswdchars = "0123456789"
        # _codeobj = _settings.reset_pwd
        # _code = "".join([pswdchars[x] for x in _codeobj if int(x) < 0x0A])
        # val = RadioSettingValueString(3, 6, _code, False)
        # val.set_charset(pswdchars)
        # rs = RadioSetting("reset_pwd", "Reset Password", val)
        # def apply_pwd(setting, obj):
        #     value = []
        #     for j in range(0, 6):
        #         try:
        #             value.append(pswdchars.index(str(setting.value)[j]))
        #         except IndexError:
        #             value.append(0xFF)
        #     obj.reset_pwd = value
        # rs.set_apply_callback(apply_pwd, _settings)
        # key_grp.append(rs)


        # Key Settings
        #
        _msg = str(_settings.dispstr).split("\0")[0]
        val = RadioSettingValueString(0, 15, _msg)
        val.set_mutable(True)
        rs = RadioSetting("dispstr", "Display Message - Interface Display Edit", val)
        key_grp.append(rs)

        dtmfchars = "0123456789"
        _codeobj = _settings.ani_code
        _code = "".join([dtmfchars[x] for x in _codeobj if int(x) < 0x0A])
        val = RadioSettingValueString(3, 6, _code, False)
        val.set_charset(dtmfchars)
        rs = RadioSetting("ani_code", "ANI Code", val)
        def apply_ani_id(setting, obj):
            value = []
            for j in range(0, 6):
                try:
                    value.append(dtmfchars.index(str(setting.value)[j]))
                except IndexError:
                    value.append(0xFF)
            obj.ani_code = value
        rs.set_apply_callback(apply_ani_id, _settings)
        key_grp.append(rs)

        rs = RadioSetting("pf1_shrt", "PF1 SHORT Key function",
                          RadioSettingValueList(
                              PFKEYSHORT_LIST,
                              PFKEYSHORT_LIST[_settings.pf1_shrt]))
        key_grp.append(rs)
        rs = RadioSetting("pf1_long", "PF1 LONG Key function",
                          RadioSettingValueList(
                              PFKEYLONG_LIST,
                              PFKEYLONG_LIST[_settings.pf1_long]))
        key_grp.append(rs)
        rs = RadioSetting("pf2_shrt", "PF2 SHORT Key function",
                          RadioSettingValueList(
                              PFKEYSHORT_LIST,
                              PFKEYSHORT_LIST[_settings.pf2_shrt]))
        key_grp.append(rs)
        rs = RadioSetting("pf2_long", "PF2 LONG Key function",
                          RadioSettingValueList(
                              PFKEYLONG_LIST,
                              PFKEYLONG_LIST[_settings.pf2_long]))
        key_grp.append(rs)

#       SCAN GROUPs
        rs = RadioSetting("ScnGrpA_Act", "Scan Group A Active",
                          RadioSettingValueList(SCANGRP_LIST, SCANGRP_LIST[_settings.ScnGrpA_Act]))
        scan_grp.append(rs)
        rs = RadioSetting("ScnGrpB_Act", "Scan Group B Active",
                          RadioSettingValueList(SCANGRP_LIST, SCANGRP_LIST[_settings.ScnGrpB_Act]))
        scan_grp.append(rs)


        # rs = RadioSetting("Group_1_lower", "Scan Group 1 Lower",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_1_lower))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_1_upper", "Scan Group 1 Upper",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_1_upper))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_2_lower", "Scan Group 2 Lower",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_2_lower))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_2_upper", "Scan Group 2 Upper",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_2_upper))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_3_lower", "Scan Group 3 Lower",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_3_lower))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_3_upper", "Scan Group 3 Upper",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_3_upper))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_4_lower", "Scan Group 4 Lower",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_4_lower))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_4_upper", "Scan Group 4 Upper",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_4_upper))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_5_lower", "Scan Group 5 Lower",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_5_lower))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_5_upper", "Scan Group 5 Upper",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_5_upper))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_6_lower", "Scan Group 6 Lower",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_6_lower))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_6_upper", "Scan Group 6 Upper",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_6_upper))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_7_lower", "Scan Group 7 Lower",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_7_lower))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_7_upper", "Scan Group 7 Upper",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_7_upper))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_8_lower", "Scan Group 8 Lower",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_8_lower))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_8_upper", "Scan Group 8 Upper",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_8_upper))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_9_lower", "Scan Group 9 Lower",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_9_lower))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_9_upper", "Scan Group 9 Upper",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_9_upper))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_10_lower", "Scan Group 10 Lower",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_10_lower))
        # scan_grp.append(rs)
        # rs = RadioSetting("Group_10_upper", "Scan Group 10 Upper",
        #                   RadioSettingValueInteger(1, 999, _scan_groups.Group_10_upper))
        # scan_grp.append(rs)

        # VFO A Settings
        #
        rs = RadioSetting("work_mode_a", "VFO A Workmode",
                          RadioSettingValueList(WORKMODE_LIST, WORKMODE_LIST[_settings.work_mode_a]))
        vfoa_grp.append(rs)
        rs = RadioSetting("work_ch_a", "VFO A Work Channel",
                          RadioSettingValueInteger(1, 999, _settings.work_ch_a))
        vfoa_grp.append(rs)
        rs = RadioSetting("vfoa.rxfreq", "VFO A Rx Frequency (Hz)",
                          RadioSettingValueInteger(
                              134000000, 520000000, _vfoa.rxfreq * 10, 250))
        vfoa_grp.append(rs)
        rs = RadioSetting("vfoa.power", "VFO A Power",
                          RadioSettingValueList(
                              POWER_LIST, POWER_LIST[_vfoa.power]))
        vfoa_grp.append(rs)
        rs = RadioSetting("vfoa.iswide", "VFO A Wide/Narrow",
                          RadioSettingValueList(
                              BANDWIDTH_LIST, BANDWIDTH_LIST[_vfoa.iswide]))
        vfoa_grp.append(rs)
        rs = RadioSetting("vfoa.mute_mode", "VFO A Mute (SP Mute)",
                          RadioSettingValueList(
                              SPMUTE_LIST, SPMUTE_LIST[_vfoa.mute_mode]))
        vfoa_grp.append(rs)
        rs = RadioSetting("VFO_repeater_a", "VFO A Repeater",
                          RadioSettingValueBoolean(_settings.VFO_repeater_a))
        vfoa_grp.append(rs)

        rs = RadioSetting("vfoa.scrambler", "VFO A Descramble",
                          RadioSettingValueList(
                              SCRAMBLE_LIST, SCRAMBLE_LIST[_vfoa.scrambler]))
        vfoa_grp.append(rs)

        # rs = RadioSetting("compander", "VFO A Compander",
        #                  RadioSettingValueBoolean(_vfoa.compander))
        # vfoa_grp.append(rs)

        rs = RadioSetting("vfoa.step", "VFO A Step (kHz)",
                          RadioSettingValueList(
                              STEP_LIST, STEP_LIST[_vfoa.step]))
        vfoa_grp.append(rs)
        rs = RadioSetting("vfoa.squelch", "VFO A Squelch",
                          RadioSettingValueList(
                              LIST_10, LIST_10[_vfoa.squelch]))
        vfoa_grp.append(rs)
        # VFO B Settings
        #
        rs = RadioSetting("work_mode_b", "VFO B Workmode",
                          RadioSettingValueList(WORKMODE_LIST, WORKMODE_LIST[_settings.work_mode_b]))
        vfob_grp.append(rs)
        rs = RadioSetting("work_ch_b", "VFO B Work Channel",
                          RadioSettingValueInteger(1, 999, _settings.work_ch_b))
        vfob_grp.append(rs)
        rs = RadioSetting("vfob.rxfreq", "VFO B Rx Frequency (Hz)",
                          RadioSettingValueInteger(
                              134000000, 520000000, _vfob.rxfreq * 10, 250))
        vfob_grp.append(rs)
        rs = RadioSetting("vfob.power", "VFO B Power",
                          RadioSettingValueList(
                              POWER_LIST, POWER_LIST[_vfob.power]))
        vfob_grp.append(rs)
        rs = RadioSetting("vfob.iswide", "VFO B Wide/Narrow",
                          RadioSettingValueList(
                              BANDWIDTH_LIST, BANDWIDTH_LIST[_vfob.iswide]))
        vfob_grp.append(rs)
        rs = RadioSetting("vfob.mute_mode", "VFO B Mute (SP Mute)",
                          RadioSettingValueList(
                              SPMUTE_LIST, SPMUTE_LIST[_vfob.mute_mode]))
        vfob_grp.append(rs)
        rs = RadioSetting("VFO_repeater_b", "VFO B Repeater",
                          RadioSettingValueBoolean(_settings.VFO_repeater_b))
        vfob_grp.append(rs)

        rs = RadioSetting("vfob.scrambler", "VFO B Descramble",
                          RadioSettingValueList(
                              SCRAMBLE_LIST, SCRAMBLE_LIST[_vfob.scrambler]))
        vfob_grp.append(rs)

        #rs = RadioSetting("compander", "VFO B Compander",
        #                  RadioSettingValueBoolean(_vfob.compander))
        #vfob_grp.append(rs)
        
        rs = RadioSetting("vfob.step", "VFO B Step (kHz)",
                          RadioSettingValueList(
                              STEP_LIST, STEP_LIST[_vfob.step]))
        vfob_grp.append(rs)
        rs = RadioSetting("vfob.squelch", "VFO B Squelch",
                          RadioSettingValueList(
                              LIST_10, LIST_10[_vfob.squelch]))
        vfob_grp.append(rs)

# FM RADIO PRESETS
        rs = RadioSetting("FM_radio1", "FM1 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio1))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio2", "FM2 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio2))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio3", "FM3 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio3))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio4", "FM4 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio4))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio5", "FM5 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio5))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio6", "FM6 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio6))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio7", "FM7 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio7))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio8", "FM8 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio8))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio9", "FM9 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio9))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio10", "FM10 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio10))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio11", "FM11 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio11))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio12", "FM12 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio12))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio13", "FM13 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio13))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio14", "FM14 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio14))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio15", "FM15 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio15))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio16", "FM16 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio16))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio17", "FM17 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio17))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio18", "FM18 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio18))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio19", "FM19 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio19))
        fmradio_grp.append(rs)
        rs = RadioSetting("FM_radio20", "FM20 in MHz*10",
                          RadioSettingValueInteger(760, 1080, _settings.FM_radio20))
        fmradio_grp.append(rs)

# Freq Limits settings
        #
        rs = RadioSetting("vhf_limits.rx_start", "VHF RX Lower Limit (Hz)",
                          RadioSettingValueInteger(
                              10000000, 299999999,
                              self._memobj.vhf_limits.rx_start * 10, 5000))
        lmt_grp.append(rs)
        rs = RadioSetting("vhf_limits.rx_stop", "VHF RX Upper Limit (Hz)",
                          RadioSettingValueInteger(
                              10000000, 299999999,
                              self._memobj.vhf_limits.rx_stop * 10, 5000))
        lmt_grp.append(rs)
        # rs = RadioSetting("vhf_limits.tx_start", "VHF TX Lower Limit",
        #                   RadioSettingValueInteger(
        #                       10000000, 299999999,
        #                       self._memobj.vhf_limits.tx_start * 10, 5000))
        # val.set_mutable(False)
        # vhf_lmt_grp.append(rs)
        # rs = RadioSetting("vhf_limits.tx_stop", "VHF TX Upper Limit",
        #                   RadioSettingValueInteger(
        #                       10000000, 299999999,
        #                       self._memobj.vhf_limits.tx_stop * 10, 5000))
        # val.set_mutable(False)
        # vhf_lmt_grp.append(rs)

        rs = RadioSetting("uhf_limits.rx_start", "UHF RX Lower Limit (Hz)",
                          RadioSettingValueInteger(
                              300000000, 999999999,
                              self._memobj.uhf_limits.rx_start * 10, 5000))
        lmt_grp.append(rs)
        rs = RadioSetting("uhf_limits.rx_stop", "UHF RX Upper Limit (Hz)",
                          RadioSettingValueInteger(
                              300000000, 999999999,
                              self._memobj.uhf_limits.rx_stop * 10, 5000))
        lmt_grp.append(rs)
        # rs = RadioSetting("uhf_limits.tx_start", "UHF TX Lower Limit",
        #                   RadioSettingValueInteger(
        #                       300000000, 999999999,
        #                       self._memobj.uhf_limits.tx_start * 10, 5000))
        # uhf_lmt_grp.append(rs)
        # rs = RadioSetting("uhf_limits.tx_stop", "UHF TX Upper Limit",
        #                   RadioSettingValueInteger(
        #                       300000000, 999999999,
        #                       self._memobj.uhf_limits.tx_stop * 10, 5000))
        # uhf_lmt_grp.append(rs)


# OEM info
        #
        def _decode(lst):
            _str = ''.join([chr(c) for c in lst
                            if chr(c) in chirp_common.CHARSET_ASCII])
            return _str

        def do_nothing(setting, obj):
            return

        _str = _decode(self._memobj.oem_info.model)
        val = RadioSettingValueString(0, 8, _str)
        val.set_mutable(True)
        rs = RadioSetting("oem_info.model", "Model / Bottom Banner", val)
        #rs.set_apply_callback(do_nothing, _settings)
        oem_grp.append(rs)
        _str = _decode(self._memobj.oem_info.oem1)
        val = RadioSettingValueString(0, 15, _str)
        val.set_mutable(False)
        rs = RadioSetting("oem_info.oem1", "OEM String 1", val)
        rs.set_apply_callback(do_nothing, _settings)
        oem_grp.append(rs)
        _str = _decode(self._memobj.oem_info.oem2)
        val = RadioSettingValueString(0, 15, _str)
        val.set_mutable(False)
        rs = RadioSetting("oem_info.oem2", "Firmware Version ??", val)
        rs.set_apply_callback(do_nothing, _settings)
        oem_grp.append(rs)
        # _str = _decode(self._memobj.oem_info.version)
        # val = RadioSettingValueString(0, 15, _str)
        # val.set_mutable(False)
        # rs = RadioSetting("oem_info.version", "Software Version", val)
        # rs.set_apply_callback(do_nothing, _settings)
        # oem_grp.append(rs)
        _str = _decode(self._memobj.oem_info.date)
        val = RadioSettingValueString(0, 15, _str)
        val.set_mutable(False)
        rs = RadioSetting("oem_info.date", "OEM Date", val)
        rs.set_apply_callback(do_nothing, _settings)
        oem_grp.append(rs)

        return group




        return group

    def get_settings(self):
        try:
            return self._get_settings()
        except:
            import traceback
            LOG.error("Failed to parse settings: %s", traceback.format_exc())
            return None

    def set_settings(self, settings):
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    if "." in element.get_name():
                        bits = element.get_name().split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = self._memobj.settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    else:
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        if self._is_freq(element):
                            setattr(obj, setting, int(element.value)/10)
                        else:
                            setattr(obj, setting, element.value)
                except Exception, e:
                    LOG.debug(element.get_name())
                    raise

    def _is_freq(self, element):
        return "rxfreq" in element.get_name() or "txoffset" in element.get_name() or "rx_start" in element.get_name() or "rx_stop" in element.get_name() or "tx_start" in element.get_name() or "tx_stop" in element.get_name()
